﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace MyCollections
{
    /// <summary>
    /// Multi-valued dictionary: Each key maps to a set of keys
    /// </summary>
    /// <typeparam name="TKey">The type of the key</typeparam>
    /// <typeparam name="TValue">The type of the values</typeparam>
    public class MultiValueDictionary<TKey, TValue> : IEquatable<MultiValueDictionary<TKey, TValue>>, IEnumerable<KeyValuePair<TKey, ISet<TValue>>>
    {
        public Func<ISet<TValue>> ValuesCollectionFactory { get; set; }

        public static Func<ISet<TValue>> DefaultFactory =
            () => new HashSet<TValue>();

        public MultiValueDictionary(Func<ISet<TValue>> valuesCollectionFactory)
        {
            ValuesCollectionFactory = valuesCollectionFactory;
        }

        public MultiValueDictionary() : this(DefaultFactory) { }

        private readonly IDictionary<TKey, ISet<TValue>> _data =
            new Dictionary<TKey, ISet<TValue>>();

        /// <summary>
        /// Returns a string representation of this MultiValueDictionary
        /// </summary>
        /// <returns>a string representation of this MultiValueDictionary</returns>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            foreach (KeyValuePair<TKey, ISet<TValue>> pair in _data)
            {
                sb.Append(pair.Key + ": " + String.Join(",", pair.Value) + "\n");
            }
            return sb.ToString();
        }

        /// <summary>
        /// The number of keys in this MultiValueDictionary
        /// </summary>
        public int Count { get { return _data.Count; } }

        /// <summary>
        /// Adds new (key, value) pair to this MultiValueDictionary
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        /// <returns>true if the (key, value) pair was added, false if it was a duplicate</returns>
        /// <exception cref="ArgumentNullException">if key or value is null</exception>
        public bool Add(TKey key, TValue value)
        {
            if (key == null) throw new ArgumentNullException("key");
            if (value == null) throw new ArgumentNullException("value");
            ISet<TValue> values;
            if (_data.ContainsKey(key))
            {
                values = _data[key];
                if (values.Contains(value)) return false;
            }
            else
            {
                //values = new HashSet<TValue>();
                values = ValuesCollectionFactory.Invoke();
                _data[key] = values;
            }
            values.Add(value);
            return true;
        }

        /// <summary>
        /// Returns the set of values associated with this key
        /// </summary>
        /// <param name="key">the key</param>
        /// <returns>the set of values associated with this key, or null if this key is not in the map</returns>
        /// <exception cref="ArgumentNullException">if key is null</exception>
        public ISet<TValue> Get(TKey key)
        {
            if (key == null) throw new ArgumentNullException("key");
            return _data.ContainsKey(key) ? _data[key] : null;
        }

        /// <summary>
        /// Removes a (key, value) pair from this map.
        /// If the set of values is now empty, the key must be removed
        /// </summary>
        /// <param name="key">the key</param>
        /// <param name="value">the value</param>
        /// <returns>true if an element was removed</returns>
        /// <exception cref="ArgumentNullException">if key or value is null</exception>
        public bool Remove(TKey key, TValue value)
        {
            if (key == null) throw new ArgumentNullException("key");
            if (value == null) throw new ArgumentNullException("value");
            if (!_data.ContainsKey(key)) return false;
            ISet<TValue> values = _data[key];
            bool removed = values.Remove(value);
            if (removed && values.Count == 0)
            {
                _data.Remove(key);
            }
            return removed;
        }

        /// <summary>
        /// Read-only property.
        /// Gets all the keys from this map
        /// </summary>
        public ICollection<TKey> Keys { get { return _data.Keys; } }

        /// <summary>
        /// Read-only property.
        /// Get all the values of this map.
        /// </summary>
        public ICollection<ISet<TValue>> Values { get { return _data.Values; } }

        /// <summary>
        /// Returns true if this map contains the specified key
        /// </summary>
        /// <param name="key">the key</param>
        /// <returns>true if this map contains the specified key, false otherwise</returns>
        public bool ContainsKey(TKey key)
        {
            return _data.ContainsKey(key);
        }

        /// <summary>
        /// Returns an enumerator over this map
        /// </summary>
        /// <returns>an enumerator over this map</returns>
        public IEnumerator<KeyValuePair<TKey, ISet<TValue>>> GetEnumerator()
        {
            return _data.GetEnumerator();
        }

        /// <summary>
        /// Try to get the values associated with the specified key
        /// </summary>
        /// <param name="key">the key</param>
        /// <param name="values">When this method returns, the value associated with the specified key, if the key is found; otherwise, the default value for the type of the value parameter. This parameter is passed uninitialized.</param>
        /// <returns>true if this map contains an element with the specified key; otherwise, false.</returns>
        public bool TryGetValue(TKey key, out ISet<TValue> values)
        {
            return _data.TryGetValue(key, out values);
        }


        /// <summary>
        /// For each key in keys add the (key, value) pair to the dictionary.
        /// </summary>
        /// <param name="keys">the keys</param>
        /// <param name="value">the value</param>
        /// <returns>the number of elements added to this map</returns>
        /// <exception cref="ArgumentNullException">if keys or value is null</exception>
        public int AddAll(ISet<TKey> keys, TValue value)
        {
            if (keys == null) throw new ArgumentNullException("keys");
            if (value == null) throw new ArgumentNullException("value");
            int howMany = 0;
            foreach (TKey key in keys)
            {
                Boolean added = Add(key, value);
                if (added) howMany++;
            }
            return howMany;
        }

        /// <summary>
        /// Remove all occurrences of this value from the map.
        /// If a value set becomes empty the associated key must be removed from the map
        /// </summary>
        /// <param name="value">the value to remove</param>
        /// <returns>the number of elements removed from this map</returns>
        /// <exception cref="ArgumentNullException">if value is nulll</exception>
        public int RemoveAll(TValue value)
        {
            if (value == null) throw new ArgumentNullException("value");
            int howMany = 0;
            IList<TKey> toBeDeleted = new List<TKey>();
            foreach (KeyValuePair<TKey, ISet<TValue>> pair in _data)
            {
                bool removed = pair.Value.Remove(value);
                if (removed)
                {
                    howMany++;
                    if (pair.Value.Count == 0)
                    {
                        //_data.Remove(pair.Key);
                        toBeDeleted.Add(pair.Key);
                    }
                }
            }
            foreach (TKey key in toBeDeleted)
            {
                _data.Remove(key);
            }
            return howMany;
        }

        /// <summary>
        /// Compares two MultiValueMap objects.
        /// Defined by the interface IEquatable<MultiValueDictionary<TKey, TValue>>
        /// </summary>
        /// <param name="other">the map to compare to</param>
        /// <returns>true if the maps have the same keys, and for each key have the same values</returns>
        public bool Equals(MultiValueDictionary<TKey, TValue> other)
        {
            // http://stackoverflow.com/questions/13470335/comparing-two-dictionaries-for-equal-data-in-c
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            if (this.Count != other.Count)
            {
                return false;
            }
            foreach (TKey key in _data.Keys)
            {
                ISet<TValue> thisValues = this.Get(key);
                ISet<TValue> otherValues = other.Get(key);
                return thisValues.SetEquals(otherValues);
                // http://msdn.microsoft.com/en-us/library/bb346516(v=vs.110).aspx
            }
            return true;
        }

        /// <summary>
        /// Compares this map to another object
        /// </summary>
        /// <param name="obj">the object to compare to</param>
        /// <returns>returns true if obj is a MultiValueDictionar which is equals to this map</returns>
        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != this.GetType()) return false;
            return Equals((MultiValueDictionary<TKey, TValue>)obj);
        }

        /// <summary>
        /// Gets the hash code of this map.
        /// If two maps are equal they must have the same hash codes.
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            //return (_data != null ? _data.GetHashCode() : 0);
            int hash = 0;
            foreach (KeyValuePair<TKey, ISet<TValue>> pair in _data)
            {
                hash += pair.GetHashCode();
                foreach (TValue value in pair.Value)
                {
                    hash += value.GetHashCode();
                }
            }
            return hash;
        }

        /// <summary>
        /// Returns a non-generic IEnumerator over this collection
        /// </summary>
        /// <returns>a non-generic IEnumerator over this collection</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}
